use std::path::{Path, PathBuf};

use album::Album;
use artist::Artist;
use mongod::{Model, Referencable, Reference};
use mongodb::bson::doc;
use serde_json::json;
use track::Track;
use walkdir::WalkDir;

pub mod album;
pub mod artist;
pub mod metadata;
pub mod playlist;
pub mod track;
pub mod user;

/// Checks if a file has a music file extension
fn is_music_file(path: &Path) -> bool {
    if let Some(extension) = path.extension() {
        match extension.to_str().unwrap_or("").to_lowercase().as_str() {
            "mp3" | "flac" | "wav" | "aac" | "ogg" | "m4a" | "opus" => return true,
            _ => return false,
        }
    }
    false
}

pub struct Libary {
    root_dir: PathBuf,
}

impl Libary {
    pub async fn new(root_dir: PathBuf) -> Self {
        Self { root_dir }
    }

    pub async fn find_or_create_artist(&self, artist: &str) -> Reference {
        if let Some(artist) = Artist::find_one(doc! { "name": artist }).await {
            artist.reference()
        } else {
            Artist::create(artist).await.reference()
        }
    }

    pub async fn find_or_create_album(&self, album: &str, artist_id: Option<&str>) -> Reference {
        if let Some(album) = Album::find_one(doc! { "title": album, "artist_id": artist_id}).await {
            album.reference()
        } else {
            Album::create(album, artist_id).await.reference()
        }
    }

    pub async fn add_path_to_library(&self, path: &str) {
        // search for path already present
        if let Some(_) = Track::find_one_partial(doc! { "path": path }, json!({})).await {
            return;
        }

        // add track to library
        let metadata = metadata::get_metadata(path);

        let mut entry = json!({
            "_id": uuid::Uuid::new_v4().to_string(),
            "path": path,
        });

        if let Some(meta) = &metadata {
            if let Some(artist) = meta.artist() {
                let artist_id = self.find_or_create_artist(artist).await;
                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("artist_id".into(), artist_id.into());
            }

            if let Some(album) = meta.album() {
                let album_id = self
                    .find_or_create_album(
                        album,
                        entry
                            .as_object()
                            .unwrap()
                            .get("artist_id")
                            .map(|x| x.as_str().unwrap()),
                    )
                    .await;
                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("album_id".into(), album_id.into());
            }

            if let Some(title) = meta.title() {
                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("title".into(), title.into());
            }
        }

        if let Some(metadata) = metadata {
            entry
                .as_object_mut()
                .unwrap()
                .insert("meta".into(), metadata.0);
        }

        if entry.as_object().unwrap().get("title").is_none() {
            entry.as_object_mut().unwrap().insert(
                "title".into(),
                std::path::Path::new(path)
                    .file_stem()
                    .unwrap()
                    .to_str()
                    .unwrap()
                    .into(),
            );
        }

        Track::create(&entry.as_object().unwrap()).await;
    }

    pub async fn get_artists(&self) -> Vec<Artist> {
        Artist::find(doc! {}, None).await.unwrap()
    }

    pub async fn get_artist_by_id(&self, id: &str) -> Option<Artist> {
        Artist::get(id).await
    }

    pub async fn get_albums_by_artist(&self, artist: &str) -> Vec<Album> {
        let artist = format!("artist::{artist}");
        Album::find(doc! { "artist_id": artist}, None)
            .await
            .unwrap()
    }

    pub async fn get_album_by_id(&self, album: &str) -> Option<Album> {
        Album::get(album).await
    }

    pub async fn get_tracks_of_album(&self, album: &str) -> Vec<Track> {
        Track::find(doc! { "album_id": album}, None).await.unwrap()
    }

    pub async fn get_track_by_id(&self, track_id: &str) -> Option<Track> {
        Track::get(track_id).await
    }

    pub async fn rescan(&self) {
        for entry in WalkDir::new(self.root_dir.clone())
            .follow_links(true)
            .into_iter()
            .take(30) // todo : remove
            .filter_map(|e| e.ok())
        {
            let path = entry.path();
            if path.is_file() && is_music_file(path) {
                let path = path.to_string_lossy().to_string();

                println!("Found {path}");
                self.add_path_to_library(&path).await;
            }
        }
    }
}
